#!/bin/bash
#
# checks the unisolvence of a set of nodes as a mass-lumping basis.
#
# $Id: quadrule-ml-check.sh,v 1.12 2016/07/25 08:16:44 zlb Exp $
#

temp="quadrule-ml-check-tmp-$$"
trap "/bin/rm -fr $temp*" 0 1 2 3 15

if test "$1" = "batch"; then
    shift
    sh `dirname $0`/quadrule-getrules.sh "$@" | awk '
#---------------------------------------------------------------------------
function run_case() {
    local i
    for (i = 0; i < found; i++) {
	if (dim == 2) {
	    printf "%s, orders %d %d %d, ", prefix, o0[i], o1[i], o2[i]
	    cmd = sprintf("%s quick 2 %d %d %d", "'$0'", o0[i], o1[i], o2[i])
	} else {
	    printf "%s, orders %d %d %d %d, ", prefix,
					o0[i], o1[i], o2[i], o3[i] 
	    cmd = sprintf("%s quick 3 %d %d %d %d", "'$0'",
					o0[i], o1[i], o2[i], o3[i])
	}
	print rule | cmd; close(cmd)
    }
}
BEGIN {flag = 0}
/^[23]D .* point order/ {
    if (flag) run_case();
    tmp = $7
    gsub("[,(]","",tmp);
    # Note: $5 = quadrature order
    prefix = sprintf("%s %s point, orbits %s %s", $1, $2, tmp, $15)
    flag = 1; rule=""
    dim = substr($1, 1, 1)
    if (dim != 2 && dim != 3) {
	print "Invalid dimension: " dim ", case skipped."
	flag = 0
	next
    }
    points = $2
    order = $5
    # Find o0, o1, o2, o3 matching (dim,points,order)
    found = 0;
    if (dim == 2) {
	for (i = 1; i <= (order + 2) / 2; i++) {
	    k = order + 2 - i
	    for (j = i; j <= k; j++) {
		if (3 + 3 * (j - 1) + (k - 2) * (k - 1) / 2 == points) {
		    o0[found] = i; o1[found] = j; o2[found] = k
		    found++
		}
	    }
	}
    } else {
	for (i = 1; i <= (order + 2) / 2; i++) {
	    l = order + 2 - i
	    for (j = i; j <= l; j++)
		for (k = j; k <= l; k++) {
		    if (4+6*(j-1)+4*(k-2)*(k-1)/2+(l-3)*(l-2)*(l-1)/6==points) {
			o0[found]=i; o1[found]=j; o2[found]=k; o3[found]=l
			found++
		    }
		}
	}
    }
    if (!found) {
	printf "Cannot find parameters matching order and points, "
	printf "case skipped.\n"
	flag = 0
	next
    }
    next
}
{if (rule != "") rule = rule "\n" $0; else rule = $0}
END {if (flag) run_case()}
#---------------------------------------------------------------------------
'
    exit 0
fi

quick=false
if test "$1" = "quick"; then
    shift
    quick=true
fi

if test $# -eq 4 -a "$1" = "2" ; then
    dim="2"
    O0="$2"
    O1="$3"
    O2="$4"
    O3=""
    if test 1 -gt $O0 -o $O0 -gt $O1 -o $O1 -gt $O2; then
	echo 1>&2 "Invalid orders: $O0 $O1 $O2"
	exit 1
    fi
    quad_order=$(($O0 + $O2 - 2))
elif test $# -eq 5 -a "$1" = "3" ; then
    dim="3"
    O0="$2"
    O1="$3"
    O2="$4"
    O3="$5"
    if test 1 -gt $O0 -o $O0 -gt $O1 -o $O1 -gt $O2 -o $O2 -gt $O3; then
	echo 1>&2 "Invalid orders: $O0 $O1 $O2 $O3"
	exit 1
    fi
    quad_order=$(($O0 + $O3 - 2))
else
    cat <<END 1>&2
Usage:
    $0 [quick] 2 order0 order1 order2		# 2D
    $0 [quick] 3 order0 order1 order2 order3	# 3D
    $0 batch {... quadrule_output_files ...}	# batch
END
    exit 1
fi

# locate header file "quad-permu.h"
if test -r "quad-permu.h"; then
    INC=""
elif test -r "$HOME/phg/include/phg/quad-permu.h"; then
    INC="-I$HOME/phg/include/phg"
elif test -r "../include/phg/quad-permu.h"; then
    INC="-I../include/phg"
else
    echo 1>&2 "Cannot locate file \"quad-permu.h\", abort."
    exit 1
fi

###############################################################################
$quick || echo 1>&2 "*** Reading orbits from stdin ..."
cat <<END >$temp.c
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>

#include "quad-permu.h"

typedef long double FLOAT;
#define Pow	powl	/* matching FLOAT */
#define Fabs	fabsl	/* matching FLOAT */
#define _F(a)	a ## L	/* (FLOAT)(a) */

static FLOAT points[] = {
END
#------------------------------------------------------------------------------
if $quick; then
    # use the original input orbits
    . `dirname $0`/read_rule.func
    read_rule | grep -v 'dup' | sed -e 's/t = /    /' | \
    sed -e 's/cons\(.*)\) *$/Cons\1,/' -e 's/perm\(.*)\) *$/Perm\1,/'
else
    # call ./newton to refine the input orbits (requiring weights)
    `dirname $0`/newton $dim $quad_order 2>/dev/null | grep 'Perm\|Cons'
fi | sed -e 's/\*10\^(\([+-]*[0-9]\+\))/e\1/g' >>$temp.c
#------------------------------------------------------------------------------
cat <<END >>$temp.c
};

static int N = (sizeof(points) / sizeof(points[0]));
static int dim;

static FLOAT
monomial(int p[], FLOAT x[])
/* evaluates the monomial \Pi x[i]^p[i] at the point x */
{
    int i;
    FLOAT val = 1.0;

    for (i = 0; i <= dim; i++) {
	if (p[i] == 0)
	    continue;
	if (x[i] == 0.0)
	    return 0.0;
	val *= Pow(x[i], (FLOAT)p[i]);
    }

    return val;
}

int
main(int argc, char *argv[])
{
    int o0, o1, o2, o3, n, i, j, j1, j2, j3, k, l, m, p[4], ret;
    FLOAT *A, *x, d, r;

    assert(argc == 5 || argc == 6);
    dim = atoi(argv[1]);
    assert((dim == 2 && argc == 5) || (dim == 3 && argc == 6));
    o0 = atoi(argv[2]);
    o1 = atoi(argv[3]);
    o2 = atoi(argv[4]);
    if (dim == 3)
	o3 = atoi(argv[5]);
    else
	o3 = o2;
    assert(o0 > 0 && o0 <= o1 && o1 <= o2 && o2 <= o3);

    n = N / (dim + 1);
    //printf("Number of points = %d\n", n);
    assert(n * (dim + 1) == N);

    /* check consistency bwtween the orbits and the orders */
    if (dim == 2) {
	i = 3 + 3 * (o1 - 1) + (o2 - 2) * (o2 - 1) / 2;
    }
    else {
	i = 4 + 6 * (o1 - 1) + 4 * (o2 - 2) * (o2 - 1) / 2
		+ (o3 - 3) * (o3 - 2) * (o3 - 1) / 6;
    }
    //printf("Number of bases = %d\n", i);
    assert(i == n);

    A = malloc(n * n * sizeof(*A));
    x = A;

    /* vertices */
    for (j = 0; j < dim + 1; j++) {
	p[0] = p[1] = p[2] = p[3] = 0;
	p[j] = o0;
	for (i = 0; i < n; i++)
	    *(x++) = monomial(p, points + i * (dim + 1));
    }

    /* edges */
    j1 = j2 = 0;
    for (j = 0; j < (dim == 2 ? 3 : 6); j++) {
	if (++j2 > dim)
	    j2 = (++j1) + 1;
	/* edge (j1,j2) */
	p[0] = p[1] = p[2] = p[3] = 0;
	for (k = 1; k < o1; k++) {
	    p[j1] = k; p[j2] = o1 - k;
	    for (i = 0; i < n; i++)
	    	*(x++) = monomial(p, points + i * (dim + 1));
	}
    }

    /* faces */
    j1 = 0; j2 = j3 = 1;
    for (j = 0; j < (dim == 2 ? 1 : 4); j++) {
	if (++j3 > dim) {
	    if (++j2 > dim - 1)
		j2 = (++j1) + 1;
	    j3 = j2 + 1;
	}
	/* face (j1,j2,j3) */
	p[0] = p[1] = p[2] = p[3] = 0;
	for (k = 1; k < o2; k++)
	    for (l = 1; l < o2 - k; l++) {
		p[j1] = k; p[j2] = l; p[j3] = o2 - k - l;
		for (i = 0; i < n; i++)
		    *(x++) = monomial(p, points + i * (dim + 1));
	    }
    }

    /* volume */
    if (dim == 3) {
	for (k = 1; k < o3; k++)
	    for (l = 1; l < o3 - k; l++)
		for (m = 1; m < o3 - k - l; m++) {
		    p[0] = k; p[1] = l; p[2] = m; p[3] = o3 - k - l -m;
		    for (i = 0; i < n; i++)
			*(x++) = monomial(p, points + i * (dim + 1));
		}
    }

    ret = 0;
#define a(i,j)	(A[(i) * n + (j)])
    for (i = 0; i < n; i++) {
	d = Fabs(a(i,i));
	k = i;
	for (j = i + 1; j < n; j++) {
	    if ((r = Fabs(a(j,i))) > d) {
		d = r;
		k = j;
	    }
	}
	if (Fabs(d) < 1e-12) {
	    ret = 1;
	    break;
	}
	if (k != i) {
	    /* exchange row i with row k */
	    for (j = i; j < n; j++) {
		d = a(i,j);
		a(i,j) = a(k,j);
		a(k,j) = d;
	    }
	}
	if ((d = a(i,i)) != 1.0) {
	    d = 1.0 / d;
	    for (j = i + 1; j < n; j++)
		a(i,j) *= d;
	}
	for (j = i + 1; j < n; j++) {
	    if ((d = a(j,i)) == 0.0)
		continue;
	    for (k = i + 1; k < n; k++)
		a(j,k) -= d * a(i,k);
	}
    }

    free(A);

    return ret;
}
END
###############################################################################

$quick || echo 1>&2 "*** Compiling the test program ..."
cc -O2 -Wall -g -o $temp $INC $temp.c -lm
$quick || echo 1>&2 "*** Running the test program ..."
if ./$temp $dim $O0 $O1 $O2 $O3; then
    echo "unisolvence check [01;34mPASSED[0m"
    exit 0
else
    echo "unisolvence check [01;31mFAILED[0m"
    exit 1
fi
